import com.sun.javadoc.*;
import java.util.*;
import java.io.*;

public class InvertedIndex {
    public static final String TOC_PAGE_NAME = "InvIndex.html";
    public static final String INDEX_PAGE_NAME = "InvIndexConcord.html";
    public static final String HELP_PAGE_NAME = "InvIndexHelp.html";
    public static final String PAGE_PREFIX = "InvIndex-";
    public static final String PAGE_SUFFIX = ".html";

    public static final int NAME_LIMIT = 42;

    private HashSet nonwords;
    private TreeSet entries;
    PrintWriter tocPageOut;
    PrintWriter pageOut;
    PrintWriter indexPageOut;
    PrintWriter helpPageOut;
    int linesPer;
    String title;
    ArrayList charlist;
    String mostRecentLinkLine = "<br>";

    public InvertedIndex(String t) {
	nonwords = new HashSet();
	title = t;
	for(int nx = 0; nx < nonwordList.length; nx++) {
	    nonwords.add(nonwordList[nx]);
	}
	entries = new TreeSet();
    }

    public void process(List docs, DocErrorReporter rep) {
	for(int ix = 0; ix < docs.size(); ix++) {
	    RiEntry r = (RiEntry)(docs.get(ix));
	    // rep.printNotice(r.name + ": " + getDesc(r.doc.commentText(), r.doc));
	    processClass(r, rep);
	}
    }


    public void processClass(RiEntry r, DocErrorReporter rep) {
	StringTokenizer st = new StringTokenizer(getDesc(r.doc.commentText(), r.doc), " \t\r\n\f()[]");
	ArrayList a = new ArrayList();
	HashSet hs = new HashSet();
	int ix, wx;
	String wrd;
	for(ix = 0; st.hasMoreTokens(); ix++) {
	    wrd = st.nextToken();
	    a.add(wrd);
	}
	for(wx = 0; wx < ix; wx++) {
	    String sw = stripword((String)(a.get(wx))).toLowerCase();
	    if (sw.length() > 2 && !(nonwords.contains(sw))) {
		TEntry t1 = new TEntry(r, wx, sw, a);
		if (!hs.contains(sw)) {
		    // rep.printNotice("Using word: " + sw);
		    entries.add(t1);
		    hs.add(sw);
		}
	    }
	}
	return;
    }

    public String stripword(String s) {
	// gotta finish this
	int sp = 0;
	int len = s.length();
	int ep = len;
	char c1;
	for(sp = 0; sp < len; sp++) {
	    c1 = s.charAt(sp);
	    if (Character.isLetterOrDigit(c1)) break;
	}
	if (sp == len) return "";

	for(ep--; ep > sp; ep--) {
	    c1 = s.charAt(ep);
	    if (Character.isLetterOrDigit(c1)) break;
	}
	return s.substring(sp,ep+1);
    }

    public static final String filebase = "invIndex-";

    public boolean writeOutput(RootDoc root, File outputDir, String base, DocErrorReporter rep) {
	Iterator it;
	TEntry entry;
	char c1;

	// first, see what letters we have to work with
	charlist = new ArrayList();
	char prevchar = '@';
	char tchar;
	for(it = entries.iterator(); it.hasNext(); ) {
	    entry = (TEntry)(it.next());
	    tchar = entry.word.charAt(0);
	    if (tchar != prevchar) {
		charlist.add(new Character(tchar));
		prevchar = tchar;
	    }
	}

	try {
	    tocPageOut = new PrintWriter(new FileWriter(new File(outputDir, TOC_PAGE_NAME)));
	    indexPageOut = new PrintWriter(new FileWriter(new File(outputDir, INDEX_PAGE_NAME)));
	    helpPageOut =  new PrintWriter(new FileWriter(new File(outputDir, HELP_PAGE_NAME)));
	}
	catch (IOException ie) {
	    rep.printError("Error in opening output page: " + ie);
	    return false;
	}
	    
	// write the table-of-contents page
	writeLines(tocPageOut, headerLines);
	tocPageOut.println("<ul>");
	tocPageOut.println("<p>This is the table of contents page for the inverted index.");
	tocPageOut.println("Click on one of the letters below to jump to that part of the");
	tocPageOut.println("index.");
	tocPageOut.println("<center>");
	tocPageOut.println("<p><font size='+1'><b>");
	for(int cx = 0; cx < charlist.size(); cx++) {
	    c1 = ((Character)(charlist.get(cx))).charValue();
	    String nam = getPageFileName(c1);
	    tocPageOut.println("<a href=\"" + nam + "\">" + 
			       Character.toUpperCase(c1) + 
			       "</a> &nbsp;");
	}
	tocPageOut.println("</b></font><p>");
	tocPageOut.print("<a href=\"" + INDEX_PAGE_NAME + "\">Word Index</a>");
	tocPageOut.print("  &nbsp;&nbsp;&nbsp;");
	tocPageOut.print("<a href=\"" + HELP_PAGE_NAME + "\">Help</a>");
	tocPageOut.println("</p>");	
	tocPageOut.println("</center>");
	tocPageOut.println("<p> &nbsp; </p>");
	tocPageOut.println("This inverted index was created from the");
	PackageDoc [] pds = root.specifiedPackages();
	tocPageOut.println("following Java packages:<ul>");
	int classcnt, intcnt;
	for(int px = 0; px < pds.length; px++) {
	    tocPageOut.print("<li><a href='" +
			       linkFor(pds[px], base) +
			       "'><tt>");
	    tocPageOut.println(pds[px].name() + "</tt></a> &nbsp;<small>");
	    classcnt = pds[px].allClasses().length;
	    intcnt = pds[px].interfaces().length;
	    tocPageOut.print("(" + (classcnt - intcnt) + 
			       (((classcnt - intcnt)==1)?(" class"):(" classes")) +
			       ", " + intcnt + 
			       ((intcnt==1)?(" interface"):(" interfaces")) +
			       ")</small>");
            tocPageOut.println("</li>");
	}
	tocPageOut.println("</ul>");
	tocPageOut.println("</ul>");
	writeLines(tocPageOut, footerLines);
	tocPageOut.close();

	// write out the help file
	writeLines(helpPageOut, headerLines);
	writeLines(helpPageOut, helpLines);
	writeLines(helpPageOut, footerLines);
	helpPageOut.close();

	// write the beginning of the index page
	writeLines(indexPageOut, headerLines);
	indexPageOut.println("<p> &nbsp; </p>");
	indexPageOut.println("<h2>Word Index</h2>");
	indexPageOut.println("<ul><p>");
	indexPageOut.println("This page shows each word found in the inverted index, along");
	indexPageOut.println("with the classes whose one-sentence description includes that");
	indexPageOut.println("word.  Click on a letter below to jump to that part of the word");
	indexPageOut.println("index.  Click on a word in the index to jump to that part of");
	indexPageOut.println("the inverted index.");
	indexPageOut.println("</p><center><p>");
	indexPageOut.println("<font size='+1'><b>");
	for(int cx = 0; cx < charlist.size(); cx++) {
	    c1 = ((Character)(charlist.get(cx))).charValue();
	    indexPageOut.println("<a href='#" + 
			       Character.toUpperCase(c1) + "'>" +
			       Character.toUpperCase(c1) + "</a> &nbsp;");
	}
	indexPageOut.println("</b></font><p>");
	indexPageOut.print("<a href=\"" + TOC_PAGE_NAME + "\">Main Page</a>");
	indexPageOut.print("  &nbsp;&nbsp;&nbsp;");
	indexPageOut.print("<a href=\"" + HELP_PAGE_NAME + "\">Help</a>");
	indexPageOut.println("</p></center>");	
	indexPageOut.println("<p> &nbsp; </p>");
	indexPageOut.println("This word index was created from the");
	indexPageOut.println("following Java packages:<ul>");
	for(int px = 0; px < pds.length; px++) {
	    indexPageOut.print("<li><a href='" +
			       linkFor(pds[px], base) +
			       "'><tt>");
	    indexPageOut.println(pds[px].name() + "</tt></a> &nbsp;<small>");
	    classcnt = pds[px].allClasses().length;
	    intcnt = pds[px].interfaces().length;
	    indexPageOut.print("(" + (classcnt - intcnt) + 
			       (((classcnt - intcnt)==1)?(" class"):(" classes")) +
			       ", " + intcnt + 
			       ((intcnt==1)?(" interface"):(" interfaces")) +
			       ")</small>");
            indexPageOut.println("</li>");
	}
	indexPageOut.println("</ul>");
	indexPageOut.println("<p> &nbsp; </p>");

	// write the individual pages 
	int wx;
	String linkdest;
	String prevword = "";
	for(it = entries.iterator(); it.hasNext(); ) {
	    entry = (TEntry)(it.next());
	    setPageOut(entry, outputDir, base, rep);
	    if ((linesPer % 100) == 0) {
		if (linesPer != 0) 
		    pageOut.println("</table><br>\n");
		pageOut.println("<table border=1 cellspacing=1 cellpadding=2 width='98%'>");
	    }
	    pageOut.println("<tr><td width=120 valign=top nowrap><small><code>");
	    if (!(entry.getWord().equalsIgnoreCase(prevword))) {
		char wc, pwc;
		pageOut.println("<a name='" + entry.getWord() + "'></a>");
		if (prevword.length() > 0) {
		    indexPageOut.println("</ul>");
		    pwc = Character.toUpperCase(prevword.charAt(0));
		}
		else pwc = (char)0;
		wc = Character.toUpperCase(entry.getWord().charAt(0));
		if (pwc != wc) {
		    if (pwc != (char)0) 
			indexPageOut.println("</ul>");
		    indexPageOut.println("<h1><a name='" + wc + "'></a>" 
					 + wc + "</h1>");
		    indexPageOut.println("<ul>");
		}
		indexPageOut.println("<p><font size='+1'><b>" +
				     "<a href='" + getPageFileName(wc) +
				     "#" + entry.getWord() + "'>" + 
				     entry.getWord() +
				     "</a></b></font><ul>");
		prevword = entry.getWord();
	    }
	    indexPageOut.println("<br><a href='" +
				 linkFor(entry,base) + 
				 "'>" + entry.formatNameFull() +
				 "</a>");
	    
	    linkdest = linkFor(entry, base);
	    pageOut.println("<a href='" + linkdest + "'>" +
			    entry.formatName() + "</a>");
	    pageOut.println("</code></small></td>");
	    
	    pageOut.println("<td width=220 align=right valign=top><small>");
	    for(wx = 0; wx < entry.whichword; wx++) {
		if (wx > 0) pageOut.print(" ");
		pageOut.print(reverseIndexDoclet.encode((String)(entry.words.get(wx))));
	    }
	    if (wx == 0) pageOut.print(" &nbsp; ");
	    pageOut.println("</small></td>");

	    pageOut.println("<td valign=bottom align=left><small>");
	    for( ; wx < entry.words.size(); wx++) {
		pageOut.print(" ");
		if (wx == entry.whichword) {
		    pageOut.print("<b>");
		    pageOut.print(reverseIndexDoclet.encode((String)(entry.words.get(wx))));
		    pageOut.print("</b>");
		}
		else {
		    pageOut.print(reverseIndexDoclet.decode((String)(entry.words.get(wx))));
		}
	    }
	    pageOut.println("</small></td>");
	    pageOut.println("</tr>");
	    linesPer++;
	}
	setPageOut(null, outputDir, base, rep);
	
	// write the end of the index page
	indexPageOut.println("</ul>");
	indexPageOut.println("</ul>");
	writeLines(indexPageOut, footerLines);

	// all done
	indexPageOut.close();
	return true;
    }

    // utility functions for various stuff

    /**
     * 
     */
    public static final String delimSet = " \t\n<>\r";

    public String getDesc(String ct, Doc doc) {
	StringTokenizer st;
	String clname;
	String ctd = reverseIndexDoclet.decode(ct);
	int px = doc.name().lastIndexOf('.');
	if (px > 0) clname = doc.name().substring(px + 1);
	else clname = doc.name();
	if (ctd.indexOf(clname) < 0) {
	    ctd = clname + ": " + ctd;
	}
	st = new StringTokenizer(ctd, delimSet, true);
	StringBuffer bf = new StringBuffer();
	boolean intag = false;
	boolean inwht = false;
	while(st.hasMoreTokens()) {
	    String token = st.nextToken();
	    char c1 = token.charAt(0);
	    if (c1 == '<') {
		intag = true;
	    }
	    else if (c1 == '>') {
		intag = false;
	    }
	    else if (!intag) {
		if (c1 == '\n' || c1 == '\t' || c1 == '\r') {
		    token = " "; c1 = ' ';
		}
		if (token.length() == 1 && Character.isWhitespace(c1)) {
		    if (!inwht) bf.append(" ");
		    else inwht = true;
		}
		else {
		    bf.append(token);
		    if (token.endsWith(".")) break;
		    inwht = false;
		}
	    }
	}
	return bf.toString();
    }

    public String getPageFileName(char c) {
	return PAGE_PREFIX + Character.toLowerCase(c) + PAGE_SUFFIX;
    }

    String lastPageOutName = null;
    public boolean setPageOut(TEntry e, File outputDir, String base, 
			      DocErrorReporter rep) {
	boolean ret = true;
	File nf = null;
	StringBuffer lbuf;
	if (e != null) {
	    String nam = getPageFileName(e.word.charAt(0));
	    if (lastPageOutName != null && nam.equals(lastPageOutName)) 
		return ret;
	    lastPageOutName = nam;
	    nf = new File(outputDir, lastPageOutName);
	}
	try {
	    if (pageOut != null) {
		pageOut.print("</table><br><ul>");
		pageOut.println(mostRecentLinkLine);
		pageOut.println("<p><a href='#top'>Back to top</a></p>");
		pageOut.println("</ul>");
		writeLines(pageOut, footerLines);
		pageOut.close();
	    }
	    if (nf != null) {
		pageOut = new PrintWriter(new FileWriter(nf));
		writeLines(pageOut, headerLines);
		pageOut.println("<h2><a name=top></a>Inverted Index: " +
				Character.toUpperCase(e.word.charAt(0)) +
				"</h2>");
		pageOut.println("<ul>");
		int px = charlist.indexOf(new Character(e.word.charAt(0)));
		lbuf = new StringBuffer();
		if (px > 0) {
		    char pc = ((Character)(charlist.get(px - 1))).charValue();
		    lbuf.append("<a href='");
		    lbuf.append(getPageFileName(pc));
		    lbuf.append("'>Previous Page: " + 
				Character.toUpperCase(pc) + 
				"</a> &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
		    lbuf.append('\n');
		}
		lbuf.append("<a href='");
		lbuf.append(TOC_PAGE_NAME);
		lbuf.append("'>Main Page</a> &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
		if (px < (charlist.size() - 1)) {
		    char pc = ((Character)(charlist.get(px + 1))).charValue();
		    lbuf.append("<a href='");
		    lbuf.append(getPageFileName(pc));
		    lbuf.append("'>Next Page: " + 
				    Character.toUpperCase(pc) + 
				    "</a>");
		    lbuf.append('\n');
		}
		mostRecentLinkLine = lbuf.toString();
		pageOut.println(mostRecentLinkLine);
		pageOut.println("</ul>");
	    }
	    linesPer = 0;
	}
	catch (IOException ie) {
	    rep.printError("Could not open file: " + ie);
	    ret = false;
	}
	return ret;
    }

    public String linkFor(TEntry entry, String base) {
	// this is how it seems to work for JDK 1.2
	String ret;
	if (((ProgramElementDoc)(entry.doc)).containingClass() != null) {
	    // assume inner class nesting never more than 1 deep
	    // (this is probably not a good way to do this...)
	    ret = base + "/" + entry.name.replace('.','/') + ".html";
	    int rix = ret.lastIndexOf('/');
	    ret = ret.substring(0,rix) + "." + ret.substring(rix+1);
	}
	else {
	    ret = base + "/" + entry.name.replace('.','/') + ".html";
	}
	return ret;
    }

    public String linkFor(PackageDoc doc, String base) {
	// this is how it seems to work for JDK 1.2
	String ret;
	ret = base + "/" + doc.name().replace('.','/') + "/package-summary.html";
	return ret;
    }

    public void writeLines(PrintWriter pw, String [] lines) {
	int tp;
	for(int ix = 0; ix < lines.length; ix++) {
	    tp = lines[ix].indexOf("%t");
	    if (tp >= 0) {
		pw.print(lines[ix].substring(0,tp));
		pw.print(title);
		pw.println(lines[ix].substring(tp+2));
	    }
	    else {
		pw.println(lines[ix]);
	    }
	}
	return;
    }

    // header and footer for all pages
    static String headerLines[] = {
	"<!doctype html public \"-//w3c//dtd html 4.0 transitional//en\">",
	"<html>",
	"<head>",
	"<meta http-equiv=\"Content-Type\" content=\"text/html; charset=iso-8859-1\">",
	"<meta name=\"GENERATOR\" content=\"Java Inverted Index Doclet 0.1\">",
	"<title>%t Doc Inverted Index</title>",
	"</head>",
	"<body bgcolor=\"white\">",
	"<table BORDER=0 CELLSPACING=4 CELLPADDING=3 WIDTH=\"98%\" BGCOLOR=\"#3366FF\" >",
	"<tr>",
	"<td><b><font face=\"Arial, Helvetica\"><font color=\"#FFFFFF\"><font size=+2>%t Class",
	"Documentation Inverted Index</font></font></font></b></td>",
	"</tr>",
	"</table>"
    };


    static String footerLines[] = {
	"<p> &nbsp; </p>",
	"<table BORDER=0 CELLSPACING=4 CELLPADDING=3 WIDTH=\"98%\" BGCOLOR=\"#3366FF\" >",
	"<tr>",
	"<td> &nbsp; </td>",
	"</tr>",
	"</table><ul>",
	"<font size=-1>This page generated by the ",
	reverseIndexDoclet.nameAndVersion,
	"</font>",
	"<br><font size=-1>For more information, or to obtain a copy of the doclet,",
	"mail to <a href=\"mailto:ziring@home.com\">Neal Ziring</a>.</font>",
	"</ul><br>",
	"<ul>",
	"<p><font size=-1>Java, Java 2D, and JDBC are a trademarks or registered trademarks of Sun Microsystems, Inc. in the US and other countries.</font></p>",
	"</ul>",
	"</body>",
	"</html>"
    };    

    static String helpLines[] = {
	"<center>",
	"Back to: &nbsp;&nbsp;<a href='InvIndex.html'>Main Page</a> &nbsp;&nbsp; &nbsp;&nbsp;",
	"<a href='InvIndexConcord.html'>Word Index</a>",
	"</center>",
	"<p>",
	"<h1>Doclet Inverted Index FAQ</h1>",
	"<ul>",
	"<h2>How Do I Use the Inverted Index?</h2>",
	"<p>",
	"The inverted index for the %t class documentation has been ",
	"created to help you find classes that are related to ",
	"certain concepts.  In particular, you can look up a word",
	"in the inverted index, and find all the classes whose",
	"description mention that word.",
	"<p>",
	"For example, let's assume you need to find classes",
	"that have something to do with building raster images.  You",
	"would go to the <b>R</b> section of the inverted index,",
	"and scroll down looking for the word 'raster' in boldface",
	"type.  Alternatively, you could use the word list page",
	"to jump right to the correct section of the inverted index.",
	"<p>",
	"<h2>Where Did this Idea Come From?</h2>",
	"<p>",
	"The inverted index was used for Unix documentation",
	"since the days of Unix Version 6, or perhaps even",
	"earlier.  It provides a simple and easy way to",
	"do keyword searching, without the latency of a",
	"true search engine.",
	"<p>",
	"<h2>How Does the Inverted Index Link to Real %t Documentation?</h2>",
	"<p>",
	"At the time the inverted index was generated, the base URL of",
	"a full %t documentation set was supplied.  This base URL is",
	"used to compute a full URL for the class documentation for",
	"all the %t classes and package in the HTML that the index",
	"generator doclet writes out.",
	"<p>",
	"<h2>Can I Make My Own Inverted Indices?</h2>",
	"<p>",
	"Sure!  This inverted index was generated by a simple",
	"Javadoc Doclet.  If you have the JDK 1.2.1 or later, then",
	"you have a version of Javadoc that can use this doclet.",
	"Mail to <a href='mailto:ziring@home.com'>Neal Ziring</a>",
	"to obtain a copy of the doclet, or search for the ",
	"term 'Inverted Index Doclet' in your favorite Java-oriented",
	"resource directory.",
	"</ul>",
	"<center>",
	"Back to: &nbsp;<a href='InvIndex.html'>Main Page</a>&nbsp;&nbsp;&nbsp;&nbsp;",
	"<a href='InvIndexConcord.html'>Word Index</a>",
	"</center>",
    };


    // list of words to ignore.
    static final String [] nonwordList = {
	"a", "A", 
	"all", "All",
	"among", "Among",
	"an", "An",
	"and", "And", 
	"are", "Are", 
	"as", "As", 
	"attempt", "Attempt",
	"attempting", "Attempting",
	"be", "Be", 
	"between", "Between",
	"by", "By", 
	"called", "Called", 
	"can", "Can", 
	"cannot", "Cannot", 
	"class",
	"common", "Common",
	"could", "Could", 
	"defines", "Defines",
	"defined", "Defined",
	"define", "Define",
	"did", "Did",
	"did't", "Didn't",
	"does", "Does", 
	"doesn't", "doesn'T", 
	"each", "Each", 
	"etc", "Etc",
	"even",
	"false", "False",
	"first", "First", 
	"for", "For", 
	"from", "From", 
	"get", "Get",
	"give", "Give",
	"gives", "Gives",
	"go", "Go",
	"gone", "Gone",
	"has", "Has",
	"have", "Have",
	"had", "Had",
	"haven't" , "Haven't",
	"i.e.", "I.e.",
	"if", "If", 
	"implement", "Implement", 
	"in", "In", 
	"io", "IO", "I/O",
	"into", "Into",
	"is", "Is", 
	"it", "It", 
	"its", "Its", 
	"last", "Last", 
	"least", "Least", 
	"may", "May",
	"most", "Most", 
	"must", "Must", 
	"not", "Not", 
	"of", "Of", 
	"on", "On", 
	"one", "One",
	"onto", "Onto", 
	"or", "Or", 
	"over", "Over",
	"represents", "Represents",
	"such", "Such",
	"so", "So",
	"see", "See",
	"some",
	"support", "Support",
	"supports", "Supports",
	"take", "Take",
	"takes", "Takes",
	"that", "That",
	"the", "The", 
	"them", "Them", 
	"then", "Then", 
	"there", "There", 
	"these", "These", 
	"this", "This",
	"those", "Those", 
	"though", "Through",
	"to", "To", 
	"too", "Too", 
	"true", "True",
	"two", "Two",
	"type", "Type", 
	"under", "Under", 
	"use", "Use", 
	"used", "Used", 
	"uses", "Uses", 
	"useful", "Useful", 
	"usefully", "Usefully", 
	"via", "Via", 
	"was", "Was",
	"wasn't", "Wasn't",
	"well", "Well", 
	"well-known", "Well-known", 
	"what", "What", 
	"what's", "What's", 
	"when", "When", 
	"whenever", "Whenever", 
	"where", "Where", 
	"which", "Which",
	"while", "While",
	"who", "Who", 
	"whose", "Whose", 
	"will", "Will",
	"won't", "Won't",
	"why", "Why", 
	"with", "With",
	"you", "You",
	"your", "Your",
    };

}
